# Library to load data  from data folder or yfinance

import yfinance as yf
import pandas as pd

class LoadData:
    """
    Few Important Tickers are :

    # Indices
    | ^GSPC | S&P500 | USD |
    | ^NSEI | Nifty 50 | INR |
    | ^NSEBANK | NiftyBank | INR |
    | ^BSESN | Sensex | INR |

    # Crypto
    | BTC-USD | Bitcoin | USD |
    | ETH-USD | Ethereum | USD |
    | LTC-USD | Litecoin | USD |
    | DOGE-USD | Dogecoin | USD |

    # Commodities
    | GC=F | Gold | USD |
    | SI=F | Silver | USD |
    | CL=F | Crude Oil | USD |
    | HG=F | Copper | USD |
    | ZC=F | Corn | USD |

    # US treasury
    | ^IRX | 13 Week Treasury Bill | USD |
    | ^FVX | 5 Year Treasury Note | USD |
    | ^TNX | 10 Year Treasury Note | USD |
    | ^TYX | 30 Year Treasury Bond | USD |

    # Indian NSE (National Stock Exchange)
    | RELIANCE.NS | Reliance Industries | INR |
    | TCS.NS | Tata Consultancy Services | INR | 
    | HDFCBANK.NS | HDFC Bank | INR |

    # Indian BSE (Bombay Stock Exchange)
    | RELIANCE.BO | Reliance Industries | INR |
    | TCS.BO | Tata Consultancy Services | INR |
    | HDFCBANK.BO | HDFC Bank | INR |
    """



    def __init__(self) -> None:
        pass

    def get_ticker_names(self, exchange):
        """
        Parameters
        ----------
        exchange : str
            Exchange to get the ticker name from

        Returns
        -------
        DataFrame:
            Return the DataFrame of the Ticker and Company name listed on the specified exchanged
        """

        url_nse = "https://archives.nseindia.com/content/equities/EQUITY_L.csv"
        url_bse = "https://www.bseindia.com/corporates/List_Scrips.html"
        url_nasdaq = "https://www.nasdaq.com/market-activity/stocks/screener"

        if exchange.lower() == "nse":
            return pd.read_csv(url_nse)[["SYMBOL","NAME OF COMPANY"]]
        elif exchange.lower() == "bse":
            print("BSE not supported yet. You can use NSE instead or check the companies listed there for the returned url")
            return url_bse
        elif exchange.lower() == "nasdaq":
            print("Nasdaq not supported yet. You can use NSE instead or check the companies listed there for the returned url")
            return url_nasdaq


    def get_company_name(self, ticker, exchange):
        """
        Get the company name from the ticker

        Parameters
        ----------
        ticker : str
            Ticker of the company

        exchange : str
            Exchange to get the ticker name from

        Returns
        -------
        str
            Return the company name
        """

        if exchange.lower() == "nse":
            try:
                return self.get_ticker_names(exchange).loc[self.get_ticker_names(exchange)["SYMBOL"]==ticker]["NAME OF COMPANY"].values[0]
            except:
                return "Not Found"
        elif exchange.lower() == "bse":
            try:
                return yf.Ticker(ticker).info['longName']
            except:
                return yf.Ticker(ticker).info["shortName"]
        elif exchange.lower() == "nasdaq":
            try:
                return yf.Ticker(ticker).info['longName']
            except:
                return yf.Ticker(ticker).info["shortName"]
        else:
            try:
                return yf.Ticker(ticker).info['longName']
            except:
                return yf.Ticker(ticker).info["shortName"]

        

    def load_data(self, tickers, exchange):
        """
        Load the data from the tickers

        Load Index data from nasdaq only, For Indian Index price is in INR (mostly)

        Parameters
        ----------
        tickers : list
            List of the tickers to load the data from
            
        exchange : str
            Exchange to get the ticker name from

        Returns
        -------
        dict
            Return the dictionary of the dataframes of the tickers

        if exchange = "nse" or "bse":
            Returns the historical data in INR
        if exchange = "nasdaq":
            Returns the historical data in USD

        """
        data = {}
        if exchange.lower() == "nse":
            tickers = [_+".NS" for _ in tickers]
        elif exchange.lower() == "bse":
            tickers = [_+".BO" for _ in tickers]
        elif exchange.lower() == "nasdaq":
            tickers = tickers

        for ticker in tickers:
            print(f"{ticker} Downloading...")
            data_ = yf.download(ticker, period="max")
            if data_.shape[0] !=0:
                data_ = data_.dropna()
                data.update({ticker:data_})
            else:
                print(f"{ticker} Ticker Not Found.")
            
        return data

    def save_data(self, data, path):
        """
        Save the data to the path

        Parameters
        ----------
        data : dict
            Dictionary of the dataframes of the tickers

        path : str
            Path to save the data

        Returns
        -------
        None
        """
        for ticker in data.keys():
            data[ticker].to_csv(f"{path}/{ticker}.csv")
            print(f"{ticker} Saved.")
        


# Make a Efficient Frontier

from portfolio_optimizer.portfolio import Portfolio
import plotly.graph_objects as go
import plotly.express as px
import plotly.offline as py
py.init_notebook_mode(connected=True)

import numpy as np
import pandas as pd

import warnings

pd.options.mode.chained_assignment = None


class CurveFitting:
    """
    A class to implement mth order polynomial regression using the least squares method.
    Use the `fit` method to fit the model. Then predict the Y values given X values using\\
    the `predict` method.
    """

    def __init__(self) -> None:
        self.beta = None
        self.stats = None

    def fit(self, X, Y, order=3):
        """
        Polynomial regression of order m using least squares method.
        Parameters
        ----------
        X : array_like
            Independent variable.
        Y : array_like
            Dependent variable.
        order : int, optional
            Order of the polynomial. Default is 3.
        Returns
        -------
        beta : array_like
            Coefficients of the polynomial regression model.
        """
        self.n = len(X)
        Xis = np.zeros(2 * order + 1)
        Yis = np.zeros(order + 1)
        for i in range(0, 2 * order + 1):
            if i == 0:
                Xis[i] = self.n
                continue
            xi = np.sum(X**i)
            Xis[i] = xi

        for i in range(1, order + 2):
            yi = np.sum(Y * (X ** (i - 1)))
            Yis[i - 1] = yi
        A = np.zeros((order + 1, order + 1))
        for i in range(0, order + 1):
            A[i] = Xis[i : i + order + 1]
        beta = np.linalg.solve(A, Yis)
        self.beta = beta
        return beta

    def predict(self, X_l):
        """
        Predict the Y values given X values.
        Parameters
        ----------
        X_l : array_like
            Independent variable.
        Returns
        -------
        Y_l : array_like
            Predicted Y values.
        """
        Y_l = np.zeros(len(X_l))
        for i in range(0, len(self.beta)):
            Y_l += self.beta[i] * X_l**i
        return Y_l


class EfficientFrontier:
    def __init__(self):
        pass

    def add_portfolio(self, portfolio:Portfolio):
        self.portfolio = portfolio

    def cal_required_parameters(self, weights, model):

        p_std = self.portfolio.portfolio_std(weights=weights)
        p_returns = self.portfolio.portfolio_expected_return(weights=weights, model=model)

        return p_std, p_returns

    def __simulate_frontier(self, model, short=False):

        n = 30000

        if short:
            weights = np.random.uniform(low=-2.0, high=2.0, size=(n, len(self.portfolio) ) )
            # weights /= np.sum(weights)
        else:
            weights = np.random.uniform(low=0.0, high=1.0, size=(n, len(self.portfolio) ) )
            # weights /= np.sum(weights)
            
            
        for _ in range(n):
            weights[_] /= np.sum(weights[_])


        p_returns = np.zeros(len(weights))
        p_risk = np.zeros(len(weights))
        sharpe_ratios = np.zeros(len(weights))
        for i in range(n):
            p_std_, p_returns_ = self.cal_required_parameters(weights[i], model)
            p_returns[i] = p_returns_
            p_risk[i] = p_std_
            sharpe_ratios[i] = (p_returns_ - self.portfolio.risk_free_rate) / p_std_


        data = pd.DataFrame(
            {
                "Risk": p_risk,
                "Returns": p_returns,
                "Sharpe ratio": sharpe_ratios,
            }
        )

        return data, weights
        

    def plot_sim(self, model, short=False):

        data , weights = self.__simulate_frontier(model, short=short)
        p_risk = data["Risk"]
        p_expected_returns = data["Returns"]
        store_weight = weights

        fig = go.Figure()
        fig.add_trace(go.Scatter(x=np.array(p_risk), y=p_expected_returns, mode='markers', name="Portfolio", customdata=store_weight))

        fig.update_layout(
            title="Efficient frontier",
            xaxis_title="Standard Deviation",
            yaxis_title="Expected Returns",
            font=dict(
                family="Courier New, monospace",
                size=18,
                color="#7f7f7f"
            ), 
            
        )

        template = "Standard deviation: %{x:.4f}%<br>Expected return: %{y:.4f}%"
        for i, col in enumerate(self.portfolio.portfolio_stocks()):
            template += f"<br>{col} weight: %{{customdata[{i}]:.4f}}"

        fig.update_traces(
            hovertemplate=template
        )

        fig.show()


    def __pre_process(self, data):
        data["Returns2"] = data["Returns"].apply(lambda x: np.round(x, 3))
        returns_unique = data["Returns2"].unique()
        min_risk = []
        index = []
        for i in returns_unique:
            min_risk.append(data[data["Returns2"] == i]["Risk"].min())
            index.append(data[data["Returns2"] == i]["Risk"].idxmin())
        min_risk = np.array(min_risk)
        index = np.array(index)
        data_new = data.iloc[index]

        data_new.sort_values(by="Returns", inplace=True)
        data_new["Returns_Change"] = data_new["Returns"].diff(1).fillna(0)
        # data_final = self.__remove_outliers(data_new, "Returns_Change")
        return data_new, index


    def __fit(self, X, Y, order=3):
        cf = CurveFitting()
        _ = cf.fit(X=X, Y=Y, order=order)
        Y_pred = cf.predict(X)

        return Y_pred


    def plot_frontier(self, short=False, model="capm"):
        """
        Creates a plot of the efficient frontier.
        Parameters
        ----------
        short : bool, optional
            Whether to allow shorting, by default False
        model : str, optional
            The model to use, by default "capm". Supported models are 'capm', 'sim'
            (If you want to use 'fff3' or 'fff5', first load the fff parameters.)
        Returns
        -------
        fig : plotly.graph_objects.Figure
        Examples
        --------
        >>> ef = EfficientFrontier(portfolio)
        >>> ef.plot_frontier()
        Notes
        -----
        You have to create a Portoflio object first. Then you need to load data. Only then you can call `plot_frontier()`.
        """



        data , weights = self.__simulate_frontier(model=model, short=short)
        data_final, _ = self.__pre_process(data=data)

        Y = data_final["Risk"]
        X = data_final["Returns"]

        Y_pred = self.__fit(X, Y, 3)

        data_final["Final_Returns"] = X
        data_final["Final_Risk"] = Y_pred
        weights_final = weights[data_final.index]
        customdata = weights_final.T * 100
        sharpe_ratios = data_final["Sharpe ratio"].astype(float).values

        tempelate = "Risk Deviation: %{x:.4f}<br>Expected Returns: %{y:.4f}"
        for i, name in enumerate(self.portfolio.portfolio_stocks()):
            tempelate += f"<br>{name}: %{{customdata{[i]}:.4f}}"

        fig = px.scatter(
            x=data_final["Final_Risk"],
            y=data_final["Final_Returns"],
            labels={"x": "Standard Deviation (Risk)", "y": "Expected Returns"},
            custom_data=customdata,
            color=sharpe_ratios,
            color_continuous_scale="Viridis",
        )
        fig.update_traces(
            hovertemplate=tempelate,
            marker=dict(
                size=5,
                color=sharpe_ratios,
                colorscale="Viridis",
                showscale=True,
                colorbar=dict(title="Sharpe ratio"),
            ),
        )

        
        
        # Update the color bar name
        fig.update_coloraxes(colorbar_title_text="Sharpe ratio")

        fig.show()
        return fig




# Manages the stocks, calculate beta, alpha and their cov and all
import pandas as pd
import numpy as np
from tabulate import tabulate


class Portfolio:
    def __init__(self, freq,  benchmark={}, risk_free_rate=None) -> None:
        """
        Initializes the Portfolio class

        Parameters
        ----------
        benchmark : dict, compulsory
            Dictionary of the benchmark stock

        freq : str, compulsory
            Frequency of the data which will be further used for the calculations

        risk_free_rate : pd.DataFrame, optional
            Risk free rate of the market, by default None then taking it as 0.225

        Attributes
        -----------

        market : str
        Index representing the market, like snp500 or nifty50

        stocks : dic
        Name of the Stock --> Stock dataframe
        benchmark must be given at the time of instatiation

        market_return : Int
        Returns of the market

        merged_stocks : pd.DataFrame
        Merged dataframe of all the stocks, by default there is only benchmark

        betas : pd.DataFrame
        DataFrame of Beta of all the stocks

        alphas : pd.DataFrame
        DataFrame of Alpha of all the stocks

        returns : pd.DataFrame
        DataFrame of Returns of all the stocks

        Eg :
        apple = pd.read_csv(f"data/AAPL.csv", index_col="Date", parse_dates=True)
        stocks['APPLE'] = apple

        Returns
        -------
        None
        """
        self.market = list(benchmark.keys())[0]
        self.stocks = benchmark
        self.freq = freq
        if risk_free_rate is None:
            self.risk_free_rate = 0.225
        else:
            self.risk_free_rate = risk_free_rate["Close"].asfreq(self.freq, method="ffill").pct_change().dropna().mean()*100
        self.merged_stocks = self.stocks[self.market]
        self.market_return = self.merged_stocks.asfreq(self.freq, method="ffill").pct_change().dropna()["Close"].mean()*100 # monthly returns of the market
        self.betas = pd.DataFrame()
        self.alphas = pd.DataFrame()
        self.returns = pd.DataFrame()
        

    
    def __repr__(self) -> str:
        """
        Returns : list of stocks in the portfolio when the object is called as it is
        """
        return f"Stocks: {list(self.stocks.keys())}"

    def __str__(self):
        """
        Prints the list of the stocks in the Portfolio
        """
        return str(list(self.stocks.keys()))

    def __len__(self):
        """
        Returns the number of stocks in the Portfolio except the benchmark
        """
        return len(self.stocks)-1

    def __update(self):
        """
        Updates a stock in the dictionary

        Parameters
        ----------
        name : str
            Name of the stock
        stock : pd.DataFrame
            Stock dataframe

        Returns
        -------
        None
        """
        self.merged_stocks = self.merge_dfs()
        self.returns = self.get_returns()
        self.betas = self.cal_beta()
        self.alphas = self.cal_alpha()

    def portfolio_stocks(self ):
        return [key for key in self.stocks.keys() if key not in [self.market]]

    def merge_dfs(self, stocks=None, join="inner", columns=["Close"]):
        """
        Merges a list of dataframes into one. Uses the index as the key and `pd.concat` to merge the dataframes
        Parameters
        ----------
        stocks : dictionary
            keys : Stocks name
            values : Stocks DataFrame
        join : str, optional
            How to join the dataframes, by default "inner"
        columns : str, optional
            Which columns to merge, by default "Close"
        Returns
        -------
        pd.DataFrame
            Merged dataframe
        """
        if stocks:
            pass
        else:
            stocks = self.stocks

        df_names = stocks.keys()
        dfs = stocks.values()
        
        dfs = [df[columns] for df in dfs]
        df = pd.concat(dfs, axis=1, join=join)
        if df_names:
            df.columns = df_names
        else:
            df.columns = [f"df_{i}" for i in range(len(dfs))]

        self.merged_stocks = df
        return df

    def get_returns(self, df=None, freq="M"):

        """
        Calculates the returns of a dataframe
        Parameters
        ----------
        df : pd.DataFrame
            Dataframe to calculate returns
        Returns
        -------
        pd.DataFrame
            Returns
        """
        if df is None:
            df = self.merged_stocks
            freq = self.freq
        else:
            freq = freq

        self.returns = df.asfreq(freq=freq , method="ffill").pct_change().dropna()
        return self.returns

    def add_stock(self, stock_dic):
        """
        Adds a stock to the dictionary

        Parameters
        ----------

        stock_dic : dict
            stock_dic = {name : stock}
                name : Name of the stock
                stock : Stock dataframe

        Returns
        -------
        None
        """
        self.stocks.update(stock_dic)
        self.__update()


    
    def __setitem__(self, name, stock):
        """
        Adds a stock to the dictionary

        Parameters
        ----------
        name : str
            Name of the stock
        stock : pd.DataFrame
            Stock dataframe

        Returns
        -------
        None

        Eg :
        -------
        obj["Apple"] = apple
        """
        self.stocks[name] = stock
        self.__update()
        

    def remove_stock(self, name):
        """
        Removes a stock from the dictionary

        Parameters
        ----------
        name : str
            Name of the stock

        Returns
        -------
        None
        """
        if name in self.stocks.keys():
            self.stocks.pop(name)
            self.__update()
        else:
            print(f"{name} stock is not present in Portfolio")
        
        

    def __delitem__(self, name):
        """
        Removes a stock from the dictionary

        Parameters
        ----------
        name : str
            Name of the stock

        Returns
        -------
        None

        Eg :
        --------
        del obj["apple"]
        """
        if name in self.stocks.keys():
            self.stocks.pop(name)
            # print(f"{name} stock is removed from Portfolio")
            self.__update()
        else:
            print(f"{name} stock is not present in Portfolio")

    
    def __getitem__(self, name):
        """
        Returns the stock dataframe

        Parameters
        ----------
        name : str
            Name of the stock

        Returns
        -------
        pd.DataFrame
            Stock dataframe

        Eg :
        --------
        obj["apple"]
        """
        if name in self.stocks.keys():
            return self.stocks[name]
        else:
            print(f"{name} stock is not present in Portfolio")


    

    def cal_beta(self, df=None, benchmark="snp500", freq="M"):
        """
        Calculates the alpha and beta of a dataframe
        Parameters
        ----------
        df : pd.DataFrame
            Merged Dataframe to calculate alpha and beta
        benchmark : pd.DataFrame
            Benchmark to calculate alpha and beta, by default "snp500"
        Returns
        -------
        pd.DataFrame
            Beta
        """
        if df is None:
            df = self.merged_stocks
            returns = self.returns
            benchmark = self.market
            freq = self.freq
        else:
            freq = freq
            returns = self.get_returns(df, freq=freq)
            benchmark = benchmark
            
        betas = pd.DataFrame()
        for col in returns.columns:
            if col != benchmark:
                betas.loc[col, "beta"] = returns[col].corr(returns[benchmark]) * (returns[col].std()/returns[benchmark].std())
        
        self.betas = betas
        return betas

    def cal_alpha(self, df=None, benchmark="snp500", freq="M"):
        """
        Calculates the alpha and beta of a dataframe
        Parameters
        ----------
        df : pd.DataFrame
            Merged Dataframe to calculate alpha and beta
        benchmark : pd.DataFrame
            Benchmark to calculate alpha and beta, by default "snp500"
        Returns
        -------
        pd.DataFrame
            Alpha
        """
        if df is None:
            df = self.merged_stocks
            returns = self.returns
            betas = self.betas
            benchmark = self.market
            freq = self.freq
        else:
            freq = freq
            returns = self.get_returns(df, freq=freq)
            betas = self.cal_beta(df, benchmark=benchmark, freq=freq)
            benchmark = benchmark
            

        alphas = pd.DataFrame()
        for col in returns.columns:
            if col != benchmark:
                alphas.loc[col, "alpha"] = returns[col].mean() - betas.loc[col, "beta"] * returns[benchmark].mean()

        self.alphas = alphas
        return alphas
    
    def cov_matrix(self, df=None, freq="M"):
        """
        Calculates the covariance matrix of a dataframe
        Parameters
        ----------
        df : pd.DataFrame
            Dataframe to calculate covariance matrix
        Returns
        -------
        pd.DataFrame
            Covariance matrix
        """
        if df is None:
            df = self.merged_stocks
            returns = self.returns
            freq = self.freq
        else:
            freq = freq
            returns = self.get_returns(df, freq=freq)
        
        return returns.cov()

    def expected_returns(self, risk_free_rate=0.225, market_return=0.98,  beta=None, model="CAPM"):
        
        """
        
        Parameters
        ----------
        risk_free_rate : float
        
        
        market_return : float
        
        
        beta : pd.DataFrame, optional
            Beta of the stocks, by default None 


        model : str, optional
            Model to calculate expected returns, by default "CAPM"

        Returns
        -------
        pd.DataFrame
            Expected returns of the stocks

        Eg :
        -------
        obj.expected_returns(0.225, 0.89, beta=np.array([0.5, 0.6]))

        """

        if beta is None:
            if self.betas.shape[0] == 0:
                self.betas = self.cal_beta()
            beta = np.array(self.betas['beta']) 
            risk_free_rate = self.risk_free_rate
            market_return = self.market_return
            if self.alphas.shape[0] == 0:
                self.alphas = self.cal_alpha()
            alpha = np.array(self.alphas['alpha'])

        else:
            freq = freq

        if model.upper() == "CAPM":
            excepted_returns = risk_free_rate + np.multiply(beta, (market_return - risk_free_rate))
        elif model.upper() == "SIM":
            excepted_returns = alpha + risk_free_rate + np.multiply(beta, (market_return - risk_free_rate))
        return excepted_returns


    def portfolio_expected_return(self, weights="equals", expected_returns=None, model="CAPM"):
        """
        Calculates the portfolio return

        Parameters
        ----------
        weights : np.array
            Weights of the stocks
            or "equals" for equal weights

        expected_returns : np.array
            Expected returns of the stocks

        model : str, optional
            Model to calculate expected returns, by default "CAPM"

        Returns
        -------
        float
            Portfolio return

        Eg :
        -------
        obj.portfolio_return(weights, expected_returns)
        """
        
        if expected_returns is None:
            expected_returns = self.expected_returns(model=model)
        
        if type(weights) == str:
            weights = np.array([1/len(expected_returns)]*len(expected_returns))
        else:
            weights = weights

        return np.dot(weights, expected_returns)

    def portfolio_variance(self, weights="equals", cov_matrix=None):
        """
        Calculates the portfolio variance

        Parameters
        ----------
        weights : np.array
            Weights of the stocks
            or "equals" for equal weights

        cov_matrix : pd.DataFrame
            Covariance matrix of the stocks

        Returns
        -------
        float
            Portfolio variance

        Eg :
        -------
        obj.portfolio_variance(weights, cov_matrix)
        """
        if cov_matrix is None:
            cov_matrix = self.cov_matrix().drop(columns=self.market).drop(index=self.market).to_numpy()
        else:
            cov_matrix = cov_matrix.to_numpy()

        
        if isinstance(weights, str):
            weights = np.array([1/len(cov_matrix)]*len(cov_matrix))

        return np.dot(weights, np.dot(cov_matrix, weights))*100

    def portfolio_std(self, weights="equals", cov_matrix=None):
        """
        Calculates the portfolio standard deviation

        Parameters
        ----------
        weights : np.array
            Weights of the stocks
            or "equals" for equal weights

        cov_matrix : pd.DataFrame
            Covariance matrix of the stocks

        Returns
        -------
        float
            Portfolio standard deviation

        Eg :
        -------
        obj.portfolio_std(weights, cov_matrix)
        """

        return np.sqrt(self.portfolio_variance(weights, cov_matrix))

    def portfolio_summary(self,  expected_returns=None, cov_matrix=None):
        """
        Calculates the portfolio summary

        Parameters
        ----------
        weights : np.array
            Weights of the stocks
            or "equals" for equal weights

        expected_returns : np.array
            Expected returns of the stocks

        cov_matrix : pd.DataFrame
            Covariance matrix of the stocks

        Returns
        -------
        pd.DataFrame
            Portfolio summary

        Eg :
        -------
        obj.portfolio_summary(weights, expected_returns, cov_matrix)
        """
        if expected_returns is None:
            expected_returns = self.expected_returns()
        if cov_matrix is None:
            cov_matrix = self.cov_matrix().drop(columns=self.market).drop(index=self.market)
        else:
            cov_matrix = cov_matrix

        headers = [key for key in self.stocks.keys() if key not in [self.market]]

        print("Portfolio Summary")
        print("*****************\n")
        print(f"Stocks in the Portfolio : {headers}")
        print("**************************\n")
        
        print("Beta :")
        print("******")
        print(tabulate(self.betas.T, headers,  tablefmt="orgtbl"))
        print()


        print("Expected Returns :")
        print("******************")
        print(tabulate(pd.DataFrame(expected_returns).T, headers, tablefmt="orgtbl"))
        print()



        print("The covariance matrix is as follows")
        print("***********************************")
        print(tabulate(cov_matrix, headers=headers, tablefmt="orgtbl"))
        print()

        print(f"Portfolio Returns at equals weights: {self.portfolio_expected_return(weights='equals')}")
        print(f"Portfolio Risk at equals weights: {self.portfolio_variance(weights='equals')}")
        


# Optimize the portfolio

import numpy as np
from portfolio_optimizer.portfolio import Portfolio
from scipy.optimize import minimize

class Optimizer:

    def __init__(self) -> None:
        # self.portfolio = Portfolio(**kwargs)
        pass

    def add_portfolio(self, portfolio:Portfolio):
        self.portfolio = portfolio


    def model(self,  weights, model):

        """
        CAPM Model

        Parameters
        ----------
        weights : list
            weights of each stocks

        model : str
            Model to use for the the calculation of the expected return


        Returns
        -------
        float
            Return of the expected return of the portfolio
        """
        
        return self.portfolio.portfolio_expected_return(weights=weights, model=model)
        


    def optimize_portfolio(self, model, max_risk, short=False):


        """
        Optimize the portfolio
        

        Parameters
        ----------
        model : str
            Model to use for the optimization
        risk : float
            Risk of the portfolio
        short : bool, optional
            Allow shorting, by default False


        Returns
        -------
        Returns the optimized weights of the portfolio and maximized portfolio_returns

        """

        ini_weights = [1/len(self.portfolio)] * len(self.portfolio)

        if short:
            bounds = tuple([(-2, 2)] * len(self.portfolio))
        else:
            bounds = tuple([(0, 1)] * len(self.portfolio))

        def objective_function(weights):
                return -self.model(weights=weights, model=model)

        
        cons = (
            {'type': 'eq', 'fun': lambda x: sum(x) - 1},
            {'type': 'ineq', 'fun': lambda x: max_risk - self.portfolio.portfolio_std(weights=x)}
                )

        res = minimize(objective_function, ini_weights, method='SLSQP', bounds=bounds, constraints=cons, tol=0.0001)

        weights = res["x"]
        var = self.portfolio.portfolio_std(weights=res.x)
        portfolio_expected_return = -res.fun

        if res["success"]:
            print("Optimized successfully.")
        else:
            print(f"Optimization failed. {res['message']}")
            print("Here are the last results:")
        
        print(f"Expected Portfolio's Returns : {portfolio_expected_return:.4f}")
        print(f"Risk : {var:.4f}")

        print("Expected weights:")
        print("-" * 20)
        for i, stock in enumerate(self.portfolio.portfolio_stocks()):
            print(f"{[stock]}: {weights[i]*100:.2f}%")

        

        return {
            "weights": weights,
            "portfolio_expected_return": portfolio_expected_return,
            "portfolio_variance": var,
            "portfolio_std": np.sqrt(var),
        }
